/*
 * Author: Andrew D. Fernandes <mailto:andrew@cryptonym.com>
 *         Cryptonym Corporation <http://www.cryptonym.com>
 */

#include "ecdsa.h"

#if ECDSA_TYPE == 0

#include "affine.h"
#define point apoint
#define EC_TYPE "affine"
#define BasePoint aP
#define init_point init_apoint
#define fini_point fini_apoint
#define point_alloc apoint_alloc
#define point_free apoint_free
#define point_double apoint_double
#define point_add apoint_add
#define point_scalar_multiply apoint_scalar_multiply
#define normalize_point_if_needed(p) /* null */

#elif ECDSA_TYPE == 1

#include "projective.h"
#define point ppoint
#define EC_TYPE "projective"
#define BasePoint pP
#define init_point init_ppoint
#define fini_point fini_ppoint
#define point_alloc ppoint_alloc
#define point_free ppoint_free
#define point_double ppoint_double
#define point_add ppoint_add
#define point_scalar_multiply ppoint_scalar_multiply
#define normalize_point_if_needed normalize_ppoint

#else

#error ECDSA_TYPE must be 0 (affine) or 1 (projective)

#endif


static int n_bits;
static MPI n1, k, d, w, u1, u2;
static point kP, Q, u3, u4, v;
static byte *buffer;


void init_ecdsa() {
  n_bits = mpi_get_nbits(n);
  buffer = (byte *)malloc((n_bits+7)/8);
  n1 = mpi_alloc(0);
  mpi_sub_ui(n1,n,1);
  k = mpi_alloc(0);
  d = mpi_alloc(0);
  if ( mode ) {
    ecdsa_generate_special(d);
    fprintf(stdout,"0x");
    mpi_print(stdout,d,1);
    fprintf(stdout,"\n");
  } else {
    char s[1024];
    if ( fscanf(stdin,"%1024s",s) >=1024 ) abort();
    mpi_fromstr(d,s);
  }
  Q = point_alloc(129);
  point_scalar_multiply(Q,BasePoint,d);
  kP = point_alloc(129);
  w = mpi_alloc(0);
  u1 = mpi_alloc(0);
  u2 = mpi_alloc(0);
  u3 = point_alloc(129);
  u4 = point_alloc(129);
  v = point_alloc(129);
}


void fini_ecdsa() {
  free(buffer);
  mpi_free(n1);
  mpi_free(k);
  mpi_free(d);
  point_free(Q);
  point_free(kP);
  mpi_free(w);
  mpi_free(u1);
  mpi_free(u2);
  point_free(u3);
  point_free(u4);
  point_free(v);
}


void ecdsa_generate_special( MPI k ) {
  int i;
  while (1) {
    for ( i=0; i<(n_bits+7)/8; i++ ) buffer[i] = (byte)rand();
    mpi_set_buffer( k, buffer, (n_bits+7)/8, 0 );
    if ( mpi_test_bit( k, n_bits-1 ) ) {
      mpi_set_highbit( k, n_bits-1 );
    } else {
      mpi_set_highbit( k, n_bits-1 );
      mpi_clear_bit( k, n_bits-1 );
    }
    if( !(mpi_cmp( k, n1 ) < 0) ) continue;
    if( !(mpi_cmp_ui( k, 1 ) > 0) ) continue;
    break;
  }
}


void ecdsa_sign(MPI h, MPI r, MPI s) {
  do {
    do {
      ecdsa_generate_special(k);
      point_scalar_multiply(kP,BasePoint,k);
      normalize_point_if_needed(kP);
      mpi_fdiv_r(r,kP.x,n);
    } while ( mpi_cmp_ui(r,0) == 0 );
    mpi_invm(k,k,n);
    mpi_mul(s,d,r);
    mpi_add(s,s,h);
    mpi_fdiv_r(s,s,n);
    mpi_mulm(s,k,s,n);
  } while ( mpi_cmp_ui(s,0) == 0 );
}


int ecdsa_verify(MPI h, MPI r, MPI s) {
  if ( ! ( mpi_cmp_ui(r,0) > 0 ) ) return 0;
  if ( ! ( mpi_cmp_ui(s,0) > 0 ) ) return 0;
  if ( ! ( mpi_cmp(r,n) < 0 ) ) return 0;
  if ( ! ( mpi_cmp(r,n) < 0 ) ) return 0;
  mpi_invm(w,s,n);
  mpi_mulm(u1,h,w,n);
  mpi_mulm(u2,r,w,n);
  point_scalar_multiply(u3,BasePoint,u1);
  point_scalar_multiply(u4,Q,u2);
  point_add(v,u3,u4);
  normalize_point_if_needed(v);
  mpi_fdiv_r(v.y,v.x,n);
  if ( mpi_cmp(v.y,r) == 0 ) {
    return 1;
  } else {
    return 0;
  }
}
