/* This code is based on the MD5 algorithm code from RSA Data
	Security, Inc.  Their copyright follows.
*/

/*
 **********************************************************************
 ** Copyright (C) 1990, RSA Data Security, Inc. All rights reserved. **
 **                                                                  **
 ** License to copy and use this software is granted provided that   **
 ** it is identified as the "RSA Data Security, Inc. MD5 Message     **
 ** Digest Algorithm" in all material mentioning or referencing this **
 ** software or this function.                                       **
 **                                                                  **
 ** License is also granted to make and use derivative works         **
 ** provided that such works are identified as "derived from the RSA **
 ** Data Security, Inc. MD5 Message Digest Algorithm" in all         **
 ** material mentioning or referencing the derived work.             **
 **                                                                  **
 ** RSA Data Security, Inc. makes no representations concerning      **
 ** either the merchantability of this software or the suitability   **
 ** of this software for any particular purpose.  It is provided "as **
 ** is" without express or implied warranty of any kind.             **
 **                                                                  **
 ** These notices must be retained in any copies of any part of this **
 ** documentation and/or software.                                   **
 **********************************************************************
 */

#include <stdio.h>
#include <stdlib.h>
#include <strings.h> /* bzero() */
#include <sys/stat.h>

#include "msg_dgst_rsa.h"

/* This MD5 routine works on both little and big endian machines. */

#define ROTATE_LEFT(x, n) (((x) << (n)) | ((x) >> (32 - (n))))

#define F(x, y, z) (((x) & (y)) | ((~x) & (z)))
#define G(x, y, z) (((x) & (z)) | ((y) & (~z)))
#define H(x, y, z) ((x) ^ (y) ^ (z))
#define I(x, y, z) ((y) ^ ((x) | (~z))) 
#define FF(a, b, c, d, x, s, ac) \
  {(a) += F((b), (c), (d)) + (x) + (unsigned long)(ac); \
   (a) = ROTATE_LEFT ((a), (s)); \
   (a) += (b); \
  }
#define GG(a, b, c, d, x, s, ac) \
  {(a) += G((b), (c), (d)) + (x) + (unsigned long)(ac); \
   (a) = ROTATE_LEFT ((a), (s)); \
   (a) += (b); \
  }
#define HH(a, b, c, d, x, s, ac) \
  {(a) += H((b), (c), (d)) + (x) + (unsigned long)(ac); \
   (a) = ROTATE_LEFT ((a), (s)); \
   (a) += (b); \
  }
#define II(a, b, c, d, x, s, ac) \
  {(a) += I((b), (c), (d)) + (x) + (unsigned long)(ac); \
   (a) = ROTATE_LEFT ((a), (s)); \
   (a) += (b); \
  }

typedef union
{
  unsigned char buf[64];
  unsigned long ul[16];
} MSG;

static void process_block(unsigned long *ul_p,
                          unsigned long *hold_p);

void md5_string(unsigned char *s, unsigned int len,
                unsigned char *digest_p)
{
  unsigned char uc;
  int i;
  int ii;
  int leftover = len & 0x3f;             /* same as bytes % 64 */
  unsigned int j;
  unsigned long bits[2] = {0l, 0l};
  unsigned long hold[4];
  unsigned long long lli;
  unsigned int position;
  unsigned long long blocks  = len >> 6; /* 64 bytes per block */
  unsigned long long bit_len = len << 3;
  MSG msg;
  
  hold[0] = 0x67452301;                  /* chaining variables */
  hold[1] = 0xefcdab89;
  hold[2] = 0x98badcfe;
  hold[3] = 0x10325476;
  for(lli = 0LL, position = 0; lli < blocks; lli++, position += 64) {
    bcopy(s + position, msg.buf, 64);
    if(Endian == JNS_BIG_ENDIAN) {       /* byte swapping code */
      for(j = 0; j < 64; j += 4) {
        uc             = msg.buf[j + 3];
        msg.buf[j + 3] = msg.buf[j];
        msg.buf[j]     = uc;
        uc             = msg.buf[j + 2];
        msg.buf[j + 2] = msg.buf[j + 1];
        msg.buf[j + 1] = uc;
      }
      if((bits[0] + ((unsigned long)64 << 3)) < bits[0]) {
        bits[1]++;
      }
      bits[0] += ((unsigned long)64 << 3);
    }
    process_block(msg.ul, hold);
  }
  if(leftover != 0) {
    bcopy(s + position, msg.buf, leftover);
  }
  msg.buf[leftover] = 0x80;        /* one "bit" padding */
  if(leftover < 55) {
    bzero(msg.buf + leftover + 1, 55 - leftover);
  } else if(leftover > 55) {
    bzero(msg.buf + leftover + 1, 63 - leftover);
    if(Endian == JNS_BIG_ENDIAN) { /* byte swapping code here */
      for(j = 0; j < 64; j += 4) {
        uc             = msg.buf[j + 3];
        msg.buf[j + 3] = msg.buf[j];
        msg.buf[j]     = uc;
        uc             = msg.buf[j + 2];
        msg.buf[j + 2] = msg.buf[j + 1];
        msg.buf[j + 1] = uc;
      }
      if((bits[0] + ((unsigned long)64 << 3)) < bits[0]) {
        bits[1]++;
      }
      bits[0] += ((unsigned long)64 << 3);
    }
    process_block(msg.ul, hold);
    bzero(msg.buf, 56);
  }

  if(Endian == JNS_BIG_ENDIAN) { /* byte swapping code here */
    for(j = 0; j < 56; j += 4) {
      /* Do not swap the 8 "bit length" bytes! */
      uc             = msg.buf[j + 3];
      msg.buf[j + 3] = msg.buf[j];
      msg.buf[j]     = uc;
      uc             = msg.buf[j + 2];
      msg.buf[j + 2] = msg.buf[j + 1];
      msg.buf[j + 1] = uc;
    }
    if((bits[0] + ((unsigned long)leftover << 3)) < bits[0]) {
      bits[1]++;
    }
    /* Now put computed bit length into last 8 bytes of message. */
    msg.ul[14] = bits[0] += ((unsigned long)leftover << 3);
    msg.ul[15] = bits[1];
  } else { /* little endian */
    bcopy(&bit_len, msg.buf + 56, 8);
  }
  process_block(msg.ul, hold);
  for(i = 0, ii = 0; i < 4; i++, ii += 4) {
    digest_p[ii]     = (unsigned char)( hold[i]        & 0xFF);
    digest_p[ii + 1] = (unsigned char)((hold[i] >>  8) & 0xFF);
    digest_p[ii + 2] = (unsigned char)((hold[i] >> 16) & 0xFF);
    digest_p[ii + 3] = (unsigned char)((hold[i] >> 24) & 0xFF);
  }
}

static void process_block(unsigned long *ul_p, unsigned long *hold_p)
{
  unsigned long a = hold_p[0];
  unsigned long b = hold_p[1];
  unsigned long c = hold_p[2];
  unsigned long d = hold_p[3];

  /* round one */
  FF(a, b, c, d, ul_p[ 0],  7, 0xd76aa478); /*  1 */
  FF(d, a, b, c, ul_p[ 1], 12, 0xe8c7b756); /*  2 */
  FF(c, d, a, b, ul_p[ 2], 17, 0x242070db); /*  3 */
  FF(b, c, d, a, ul_p[ 3], 22, 0xc1bdceee); /*  4 */
  FF(a, b, c, d, ul_p[ 4],  7, 0xf57c0faf); /*  5 */
  FF(d, a, b, c, ul_p[ 5], 12, 0x4787c62a); /*  6 */
  FF(c, d, a, b, ul_p[ 6], 17, 0xa8304613); /*  7 */
  FF(b, c, d, a, ul_p[ 7], 22, 0xfd469501); /*  8 */
  FF(a, b, c, d, ul_p[ 8],  7, 0x698098d8); /*  9 */
  FF(d, a, b, c, ul_p[ 9], 12, 0x8b44f7af); /* 10 */
  FF(c, d, a, b, ul_p[10], 17, 0xffff5bb1); /* 11 */
  FF(b, c, d, a, ul_p[11], 22, 0x895cd7be); /* 12 */
  FF(a, b, c, d, ul_p[12],  7, 0x6b901122); /* 13 */
  FF(d, a, b, c, ul_p[13], 12, 0xfd987193); /* 14 */
  FF(c, d, a, b, ul_p[14], 17, 0xa679438e); /* 15 */
  FF(b, c, d, a, ul_p[15], 22, 0x49b40821); /* 16 */

  /* round two */
  GG(a, b, c, d, ul_p[ 1],  5, 0xf61e2562); /* 17 */
  GG(d, a, b, c, ul_p[ 6],  9, 0xc040b340); /* 18 */
  GG(c, d, a, b, ul_p[11], 14, 0x265e5a51); /* 19 */
  GG(b, c, d, a, ul_p[ 0], 20, 0xe9b6c7aa); /* 20 */
  GG(a, b, c, d, ul_p[ 5],  5, 0xd62f105d); /* 21 */
  GG(d, a, b, c, ul_p[10],  9, 0x02441453); /* 22 */
  GG(c, d, a, b, ul_p[15], 14, 0xd8a1e681); /* 23 */
  GG(b, c, d, a, ul_p[ 4], 20, 0xe7d3fbc8); /* 24 */
  GG(a, b, c, d, ul_p[ 9],  5, 0x21e1cde6); /* 25 */
  GG(d, a, b, c, ul_p[14],  9, 0xc33707d6); /* 26 */
  GG(c, d, a, b, ul_p[ 3], 14, 0xf4d50d87); /* 27 */
  GG(b, c, d, a, ul_p[ 8], 20, 0x455a14ed); /* 28 */
  GG(a, b, c, d, ul_p[13],  5, 0xa9e3e905); /* 29 */
  GG(d, a, b, c, ul_p[ 2],  9, 0xfcefa3f8); /* 30 */
  GG(c, d, a, b, ul_p[ 7], 14, 0x676f02d9); /* 31 */
  GG(b, c, d, a, ul_p[12], 20, 0x8d2a4c8a); /* 32 */

  /* round three */
  HH(a, b, c, d, ul_p[ 5],  4, 0xfffa3942); /* 33 */
  HH(d, a, b, c, ul_p[ 8], 11, 0x8771f681); /* 34 */
  HH(c, d, a, b, ul_p[11], 16, 0x6d9d6122); /* 35 */
  HH(b, c, d, a, ul_p[14], 23, 0xfde5380c); /* 36 */
  HH(a, b, c, d, ul_p[ 1],  4, 0xa4beea44); /* 37 */
  HH(d, a, b, c, ul_p[ 4], 11, 0x4bdecfa9); /* 38 */
  HH(c, d, a, b, ul_p[ 7], 16, 0xf6bb4b60); /* 39 */
  HH(b, c, d, a, ul_p[10], 23, 0xbebfbc70); /* 40 */
  HH(a, b, c, d, ul_p[13],  4, 0x289b7ec6); /* 41 */
  HH(d, a, b, c, ul_p[ 0], 11, 0xeaa127fa); /* 42 */
  HH(c, d, a, b, ul_p[ 3], 16, 0xd4ef3085); /* 43 */
  HH(b, c, d, a, ul_p[ 6], 23, 0x04881d05); /* 44 */
  HH(a, b, c, d, ul_p[ 9],  4, 0xd9d4d039); /* 45 */
  HH(d, a, b, c, ul_p[12], 11, 0xe6db99e5); /* 46 */
  HH(c, d, a, b, ul_p[15], 16, 0x1fa27cf8); /* 47 */
  HH(b, c, d, a, ul_p[ 2], 23, 0xc4ac5665); /* 48 */

  /* round four */
  II(a, b, c, d, ul_p[ 0],  6, 0xf4292244); /* 49 */
  II(d, a, b, c, ul_p[ 7], 10, 0x432aff97); /* 50 */
  II(c, d, a, b, ul_p[14], 15, 0xab9423a7); /* 51 */
  II(b, c, d, a, ul_p[ 5], 21, 0xfc93a039); /* 52 */
  II(a, b, c, d, ul_p[12],  6, 0x655b59c3); /* 53 */
  II(d, a, b, c, ul_p[ 3], 10, 0x8f0ccc92); /* 54 */
  II(c, d, a, b, ul_p[10], 15, 0xffeff47d); /* 55 */
  II(b, c, d, a, ul_p[ 1], 21, 0x85845dd1); /* 56 */
  II(a, b, c, d, ul_p[ 8],  6, 0x6fa87e4f); /* 57 */
  II(d, a, b, c, ul_p[15], 10, 0xfe2ce6e0); /* 58 */
  II(c, d, a, b, ul_p[ 6], 15, 0xa3014314); /* 59 */
  II(b, c, d, a, ul_p[13], 21, 0x4e0811a1); /* 60 */
  II(a, b, c, d, ul_p[ 4],  6, 0xf7537e82); /* 61 */
  II(d, a, b, c, ul_p[11], 10, 0xbd3af235); /* 62 */
  II(c, d, a, b, ul_p[ 2], 15, 0x2ad7d2bb); /* 63 */
  II(b, c, d, a, ul_p[ 9], 21, 0xeb86d391); /* 64 */

  hold_p[0] += a;
  hold_p[1] += b;
  hold_p[2] += c;
  hold_p[3] += d;
}
