/*
 * Author: Andrew D. Fernandes <mailto:andrew@cryptonym.com>
 *         Cryptonym Corporation <http://www.cryptonym.com>
 */

#include "mpi-extra.h"


#include "mpi-internal.h"
#include "longlong.h"

/*
 * To compile, we assume that you have the headers and
 * libraries from GnuPG-0.9.5 (http://www.gnupg.org).
 */

/* for various */
MPI prime, A, B, n;

/* for mpi_special_mod */
static MPI tmp, qic, q[2], r[2];


/****************
 * initialize static variables
 */
void mpi_init_extra() {
  int i;
  tmp = mpi_alloc(0);
  qic = mpi_alloc(0);
  prime = mpi_alloc(0);
  for ( i=0; i<2; i++ ) { q[i] = mpi_alloc(0); r[i] = mpi_alloc(0); }
  mpi_fromstr(prime,
     "0x800000000000000000000000000000000000000000000000000000000000005F");
  A = mpi_alloc(0);
  mpi_sub_ui(A,prime,3);
  B = mpi_alloc(0);
  mpi_fromstr(B,
     "0x689DC74C9E1AB5A76603B6832B916EBC158CFB4B66B24E40BFA5BAB9B4D00FD9");
  n = mpi_alloc(0);
  mpi_fromstr(n,
     "0x3FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFCBA2A662345600C4BAB16604693B2183");
}


/****************
 * finalize and deallocate static storage
 */
void mpi_fini_extra() {
  int i;
  for ( i=0; i<2; i++ ) { mpi_free(q[i]); mpi_free(r[i]); }
  mpi_free(prime);
  mpi_free(qic);
  mpi_free(tmp);
  mpi_free(A);
  mpi_free(B);
  mpi_free(n);
}


/****************
 * set y to x shifted 255 bits to the right
 */
void mpi_rshift255( MPI y, MPI x ) {
  
  int i, tmp;
  mpi_size_t sr ;
  int b255 = 255/BITS_PER_MPI_LIMB;
  int N = 256/BITS_PER_MPI_LIMB;

  y->sign = x->sign;
  RESIZE_IF_NEEDED(y,N+2);

  if ( x->nlimbs <= b255 ) {
    y->nlimbs = 0;
  } else {
    tmp = x->d[b255] & (1<<(BITS_PER_MPI_LIMB-1));
    sr = x->nlimbs - N;
    sr = ( sr <= 0 ? 0 : sr );
    if ( sr ) {
      for ( i=0 ; i<sr ; i++ ) y->d[i] = x->d[N+i];
      for ( i=sr; i<N  ; i++ ) y->d[i] = 0;
      mpihelp_lshift(y->d,y->d,N,1);
      if ( tmp ) y->d[0] |= 1;
      y->nlimbs = N;
      MPN_NORMALIZE(y->d,y->nlimbs);
    } else {
      if ( tmp ) {
	y->nlimbs = 1;
	y->d[0] = 1;
      } else {
	y->nlimbs = 0;
      }
    }  
  }

}


/****************
 * set y to the 255 LSBs of x
 */
void mpi_255lsb( MPI y, MPI x ) {
  int i, j, n=256/BITS_PER_MPI_LIMB;
  RESIZE_IF_NEEDED(y,n);
  j = ( (x->nlimbs) < n ? (x->nlimbs) : n );
  for ( i=0; i<j; i++ ) y->d[i] = x->d[i];
  y->d[j-1] &= ~(1<<(BITS_PER_MPI_LIMB-1));
  y->nlimbs = j;
  y->sign = x->sign;
}


/****************
 * set y to x with the 255 LSBs blanked
 */
void mpi_blank255( MPI y, MPI x ) {
  int i, n=256/BITS_PER_MPI_LIMB;
  mpi_limb_t j;
  if ( x->nlimbs < n ) {
    y->nlimbs = 0;
  } else {
    y->sign = x->sign;
    RESIZE_IF_NEEDED(y,x->nlimbs);
    y->nlimbs = x->nlimbs;
    j = x->d[n-1] & (1<<(BITS_PER_MPI_LIMB-1));
    for ( i=0; i<n-1; i++ ) y->d[i] = 0;
    y->d[n-1] = j;
    for ( i=n; i<x->nlimbs; i++ ) y->d[i] = x->d[i];
    MPN_NORMALIZE(y->d,y->nlimbs);
  }
}

/****************
 * reduce x modulo (2^255+95)
 */
#if MOD_TYPE == 0

void mpi_special_mod( MPI x ) {

  int i, j;
  const unsigned long c = 95;

  mpi_rshift255(q[0],x);
  mpi_255lsb(r[0],x);
  mpi_set(x,r[0]);
  i = 0; j = 1;
  while ( mpi_cmp_ui(q[i],0) > 0 ) {
    mpi_mul_ui(qic,q[i],c);
    mpi_rshift255(q[j],qic);
    mpi_blank255(tmp,qic);
    mpi_sub(r[j],qic,tmp);
    if ( i ) {
      mpi_add(x,x,r[j]);
      i = 0; j = 1;
    } else {
      mpi_sub(x,x,r[j]);
      i = 1; j = 0;
    }
  }
  if ( mpi_cmp_ui(x,0) < 0 ) mpi_add(x,x,prime);

}

#elif MOD_TYPE == 1

void mpi_special_mod( MPI x ) {
  mpi_fdiv_r(x,x,prime);
}

#else

#error MOD_TYPE must be 0 (special) or 1 (ordinary)

#endif
