/* FILE        : sha.c
** AUTHOR      : Jim Shapiro
** DATE        : 19 March, 1999
** DISCLAIMER  : No liability is assumed by the author for any use
**               made of this program.
** DISTRIBUTION: Any use may be made of this program, as long as the
**               clear acknowledgment is made to the author in code
**               and runtime executables.  The author retains
**               copyright.
*/

#include <stdio.h>
#include <stdlib.h>
#include <strings.h> /* bzero() */
#include <sys/stat.h>

#include "msg_dgst_rsa.h"

/* This SHA routine works on both little and big endian machines. */

#define ROTATE_LEFT(x, n) (((x) << (n)) | ((x) >> (32 - (n))))

#define F1(x, y, z) (((x) & (y)) | ((~x) & (z)))
#define F2(x, y, z) ((x) ^ (y) ^ (z))
#define F3(x, y, z) (((x) & (y)) | ((x) & (z)) | ((y) & (z)))

typedef union
{
  unsigned char buf[64];
  unsigned long ul[16];
} MSG;

static void process_block(unsigned long *ul_p, unsigned long *hold_p);


void sha_string(unsigned char *s, unsigned int len, unsigned char *digest_p)
{
  unsigned char uc;
  int i;
  int ii;
  int leftover = len & 0x3f; /* same as bytes % 64 */
  unsigned int j;
  unsigned long bits[2] = {0l, 0l};
  unsigned long hold[5];
  unsigned long long lli;
  unsigned int position;
  unsigned long long blocks  = len >> 6; /* 64 bytes per block */
  unsigned long long bit_len = len << 3;
  MSG msg;
  
  hold[0] = 0x67452301; /* chaining variables */
  hold[1] = 0xefcdab89;
  hold[2] = 0x98badcfe;
  hold[3] = 0x10325476;
  hold[4] = 0xc3d2e1f0;
  for(lli = 0LL, position = 0; lli < blocks; lli++, position += 64) {
    bcopy(s + position, msg.buf, 64);
    if(Endian == JNS_LITTLE_ENDIAN) { /* byte swapping code here */
      for(j = 0; j < 64; j += 4) {
        uc             = msg.buf[j + 3];
        msg.buf[j + 3] = msg.buf[j];
        msg.buf[j]     = uc;
        uc             = msg.buf[j + 2];
        msg.buf[j + 2] = msg.buf[j + 1];
        msg.buf[j + 1] = uc;
      }
      if((bits[0] + ((unsigned long)64 << 3)) < bits[0]) {
        bits[1]++;
      }
      bits[0] += ((unsigned long)64 << 3);
    }
    process_block(msg.ul, hold);
  }
  if(leftover != 0) {
    bcopy(s + position, msg.buf, leftover);
  }
  msg.buf[leftover] = 0x80; /* one "bit" padding */
  if(leftover < 55) {
    bzero(msg.buf + leftover + 1, 55 - leftover);
  } else if(leftover > 55) {
    bzero(msg.buf + leftover + 1, 63 - leftover);
    if(Endian == JNS_LITTLE_ENDIAN) { /* byte swapping code here */
      for(j = 0; j < 64; j += 4) {
        uc             = msg.buf[j + 3];
        msg.buf[j + 3] = msg.buf[j];
        msg.buf[j]     = uc;
        uc             = msg.buf[j + 2];
        msg.buf[j + 2] = msg.buf[j + 1];
        msg.buf[j + 1] = uc;
      }
      if((bits[0] + ((unsigned long)64 << 3)) < bits[0]) {
        bits[1]++;
      }
      bits[0] += ((unsigned long)64 << 3);
    }
    process_block(msg.ul, hold);
    bzero(msg.buf, 56);
  }
  if(Endian == JNS_LITTLE_ENDIAN) { /* byte swapping code here */
    for(j = 0; j < 56; j += 4) {
      /* Do not swap the 8 "bit length" bytes! */
      uc             = msg.buf[j + 3];
      msg.buf[j + 3] = msg.buf[j];
      msg.buf[j]     = uc;
      uc             = msg.buf[j + 2];
      msg.buf[j + 2] = msg.buf[j + 1];
      msg.buf[j + 1] = uc;
    }
    if((bits[0] + ((unsigned long)leftover << 3)) < bits[0]) {
      bits[1]++;
    }
    /* Now put computed bit length into last 8 bytes of message. */
    /* Note order of bytes! */
    msg.ul[15] = bits[0] += ((unsigned long)leftover << 3);
    msg.ul[14] = bits[1];
  } else { /* big endian */
    bcopy(&bit_len, msg.buf + 56, 8);
  }
  process_block(msg.ul, hold);
  for(i = 0, ii = 0; i < 5; i++, ii += 4) {
    digest_p[ii]     = (unsigned char)((hold[i] >> 24) & 0xFF);
    digest_p[ii + 1] = (unsigned char)((hold[i] >> 16) & 0xFF);
    digest_p[ii + 2] = (unsigned char)((hold[i] >>  8) & 0xFF);
    digest_p[ii + 3] = (unsigned char)( hold[i]        & 0xFF);
  }
}

static void process_block(unsigned long *ul_p, unsigned long *hold_p)
{
  int t;
  int tm3;
  int tm8;
  int tm14;
  int tm16;
  unsigned long a = hold_p[0];
  unsigned long b = hold_p[1];
  unsigned long c = hold_p[2];
  unsigned long d = hold_p[3];
  unsigned long e = hold_p[4];
  unsigned long w_p[80];
  unsigned long tmp;
  unsigned long k;

  bcopy(ul_p, w_p, 64);
  for(t = 16, tm3 = 13, tm8 = 8, tm14 = 2, tm16 = 0; t <= 79; t++) {
    w_p[t] = w_p[tm3++] ^ w_p[tm8++] ^ w_p[tm14++] ^ w_p[tm16++];
    w_p[t] = ROTATE_LEFT(w_p[t], 1);
  }
  /* round one */
  for(t = 0, k = 0x5a827999; t <= 19; t++) {
    tmp = ROTATE_LEFT(a, 5) + F1(b, c, d) + e + w_p[t] + k;
    e = d;
    d = c;
    c = ROTATE_LEFT(b, 30);
    b = a;
    a = tmp;
  }
  /* round two */
  for(t = 20, k = 0x6ed9eba1; t <= 39; t++) {
    tmp = ROTATE_LEFT(a, 5) + F2(b, c, d) + e + w_p[t] + k;
    e = d;
    d = c;
    c = ROTATE_LEFT(b, 30);
    b = a;
    a = tmp;
  }
  /* round three */
  for(t = 40, k = 0x8f1bbcdc; t <= 59; t++) {
    tmp = ROTATE_LEFT(a, 5) + F3(b, c, d) + e + w_p[t] + k;
    e = d;
    d = c;
    c = ROTATE_LEFT(b, 30);
    b = a;
    a = tmp;
  }
  /* round four */
  for(t = 60, k = 0xca62c1d6; t <= 79; t++) {
    tmp = ROTATE_LEFT(a, 5) + F2(b, c, d) + e + w_p[t] + k;
    e = d;
    d = c;
    c = ROTATE_LEFT(b, 30);
    b = a;
    a = tmp;
  }
  hold_p[0] += a;
  hold_p[1] += b;
  hold_p[2] += c;
  hold_p[3] += d;
  hold_p[4] += e;
}
