/*
 * Author: Andrew D. Fernandes <mailto:andrew@cryptonym.com>
 *         Cryptonym Corporation <http://www.cryptonym.com>
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <sys/times.h>
#include <unistd.h>

/* For linux-2.2 on an i686 */
#define CLK_TCK 100

int debug = 0;
int mode = -1;

#include "mpi-extra.h"
#include "affine.h"
#include "projective.h"
#include "dsa.h"
#include "ecdsa.h"


#if TIME_TEST == 0

#define dsa_sign    dsa_sign
#define dsa_verify  dsa_verify
#define MOD_NAME    "(null)"

#elif TIME_TEST == 1

#define dsa_sign    ecdsa_sign
#define dsa_verify  ecdsa_verify
#if MOD_TYPE == 0
#define MOD_NAME "special_mod"
#elif MOD_TYPE == 1
#define MOD_NAME "ordinary_mod"
#else
#error MOD_TYPE must be 0 (special) or 1 (ordinary)
#endif

#else

#error TIME_TEST must be 0 (dsa) or 1 (ecdsa)

#endif


static MPI *alloc_trial_space(int n) {
  int i;
  MPI *j;
  j = (MPI *)calloc(n,sizeof(MPI));
  for (i=0;i<n;i++) j[i] = mpi_alloc(0);
  return j;
}


static void free_trial_space(MPI *j, int n) {
  int i;
  for (i=0;i<n;i++) mpi_free(j[i]);
  free(j);
}


static void write_trials(MPI *h, MPI *r, MPI *s, int n) {
  int i, j;
  MPI *a[3];
  a[0] = h; a[1] = r; a[2] = s;
  for ( i=0; i<n; i++ ) {
    for ( j=0; j<3; j++ ) {
      fprintf(stdout," 0x");
      mpi_print(stdout,a[j][i],1);
    }
    fprintf(stdout,"\n");
  }
}


static void read_trials(MPI *h, MPI *r, MPI *s, int n) {
  int i, j;
  char a[3][1024];
  for ( i=0; i<n; i++ ) {
    j = fscanf(stdin,"%1024s %1024s %1024s",a[0],a[1],a[2]);
    if ( j >= 1024 ) abort();
    mpi_fromstr(h[i],a[0]);
    mpi_fromstr(r[i],a[1]);
    mpi_fromstr(s[i],a[2]);
  }
}


int main(int argv, char *argc[]) {
  
  byte *buf;
  int i, j, vrfy;
  double su, ss, vu, vs;
  const int nTry = 64, bufSize = 20;
  struct tms ts1, ts2, tv1, tv2;
  MPI *h, *r, *s;
  const int STDERR = 2;
  
  if ( argv != 2 ) {
    fprintf(stderr,"usage: %s [ s | v ]\n",argc[0]);
    exit(-1);
  } else {
    if ( argc[1][0] == 's' ) {
      mode = 1;
    } else if ( argc[1][0] == 'v' ) { 
      mode = 0;
    } else {
      fprintf(stderr,"usage: %s [ s | v ]\n",argc[0]);
      exit(-1);      
    }
  }

  srand(time(0)); /* linux rand() is good */

  buf = (byte *)malloc(bufSize);
  mpi_init_extra();
  init_ppoint();
  init_apoint();
  init_ecdsa();
  init_dsa();

  h = alloc_trial_space(nTry);
  r = alloc_trial_space(nTry);
  s = alloc_trial_space(nTry);
  
  if ( TIME_TEST == 0 ) {
    fprintf(stderr,"timing dsa %d\n",DSA_SIZE);
  } else if ( TIME_TEST == 1 ) {
    fprintf(stderr,"timing ecdsa %s %s\n",EC_TYPE,MOD_NAME);
  } else {
    abort();
  }

  if ( mode ) {
    
    for ( i=0; i<nTry; i++ ) {
      for ( j=0; j<bufSize; j++ ) {
	buf[j] = (byte)rand();
	mpi_set_buffer(h[i],buf,bufSize,0);
      }
    }
    
    fprintf(stderr,"   signing"); fflush(stderr);
    times(&ts1);
    for ( i=0; i<nTry; i++ ) {
      dsa_sign(h[i],r[i],s[i]);
      write(STDERR,".",1);
    }
    times(&ts2);
    fputc('\n',stderr);
    
    write_trials(h,r,s,nTry);
    
    su = 1e3*((double)ts2.tms_utime-ts1.tms_utime)/(nTry*CLK_TCK);
    ss = 1e3*((double)ts2.tms_stime-ts1.tms_stime)/(nTry*CLK_TCK);
    fprintf(stderr,"ms per signature : %5.5f (user) %5.5f (system)\n",su,ss);

  } else {
    
    read_trials(h,r,s,nTry);
    
    fprintf(stderr," verifying"); fflush(stderr);
    times(&tv1);
    for ( i=0; i<nTry; i++ ) {
      vrfy = dsa_verify(h[i],r[i],s[i]);
      write(STDERR,".",1);
      if ( !vrfy ) fprintf(stderr,"\n\nerror: failed verification %d\n\n",i);
    }  
    times(&tv2);
    fputc('\n',stderr);
    
    vu = 1e3*((double)tv2.tms_utime-tv1.tms_utime)/(nTry*CLK_TCK);
    vs = 1e3*((double)tv2.tms_stime-tv1.tms_stime)/(nTry*CLK_TCK);
    fprintf(stderr,
            "ms per verification : %5.5f (user) %5.5f (system)\n",vu,vs);
    
  }
  
  free_trial_space(h,nTry);
  free_trial_space(r,nTry);
  free_trial_space(s,nTry);
  
  fini_dsa();
  fini_ecdsa();
  fini_apoint();
  fini_ppoint();
  mpi_fini_extra();
  free(buf);

  return 0;
  
}
